#define _GNU_SOURCE
#include <stdlib.h>
#include <stdint.h>
#include <sys/ioctl.h>
#include <sound/asound.h>
#include <stdio.h>
#include <string.h>

static int fd1, fd2;

static int test(int which)
{
	struct snd_ctl_elem_info info = {};
	int ret;

	memset(&info, 0x00, sizeof(info));
	info.type = SNDRV_CTL_ELEM_TYPE_BOOLEAN;
	info.count = 1;
	info.owner = 1;
	info.access = SNDRV_CTL_ELEM_ACCESS_READWRITE |
		SNDRV_CTL_ELEM_ACCESS_TLV_READWRITE;

	sprintf(info.id.name, "Test Switch");
	ioctl(fd1, SNDRV_CTL_IOCTL_ELEM_REMOVE, &info);

	ret = ioctl(fd1, SNDRV_CTL_IOCTL_ELEM_ADD, &info);
	if (ret == -1) {
		perror("Failed to add test control");
		exit(1);
	}

	ret = ioctl(fd1, SNDRV_CTL_IOCTL_ELEM_INFO, &info);
	if (ret == -1) {
		perror("Info failed");
		exit(1);
	}

	switch (which) {
	case 0:
		printf("Trying to remove control from owner fd\n");
		ioctl(fd1, SNDRV_CTL_IOCTL_ELEM_REMOVE, &info);
		break;
	case 1:
		printf("Trying to remove control from non-owner fd\n");
		ioctl(fd2, SNDRV_CTL_IOCTL_ELEM_REMOVE, &info);
		break;
	case 2:
		printf("Trying to remove control by replacing it from owner fd\n");
		info.owner = -1; // Make the control invalid
		ioctl(fd1, SNDRV_CTL_IOCTL_ELEM_REPLACE, &info);
		break;
	case 3:
		printf("Trying to remove control by replacing it from non-owner fd\n");
		info.owner = -1; // Make the control invalid
		ioctl(fd2, SNDRV_CTL_IOCTL_ELEM_REPLACE, &info);
		break;
	}

	ret = ioctl(fd1, SNDRV_CTL_IOCTL_ELEM_INFO, &info);
	if (ret == -1) {
		// Control no longer exits
		printf("\tSuccess!\n");
		return 0;
	}

	printf("\tFailed\n");

	// Cleanup
	ioctl(fd1, SNDRV_CTL_IOCTL_ELEM_REMOVE, &info);

	return 0;
}

int main(void)
{
	int i;

	fd1 = open("/dev/snd/controlC1", 0);
	if (fd1 == -1) {
		perror("");
		exit(1);
	}

	fd2 = open("/dev/snd/controlC1", 0);
	if (fd2 == -1) {
		perror("");
		exit(1);
	}

	for (i = 0; i < 4; i++)
		test(i);	

	return 0;
}
